#!/bin/bash

shellfu import pretty
shellfu import isa

#
# YAML write utilities
#

# Ranges of YAML-invalid characters
#
# List of Perl regex ranges; must be usable in both s/[]// and tr///.
# Note that by default this list is set during shellfu init unless
# it has been set to non-empty string.
#
# Overriding might result in invalid YAML log!
#
__YAML__BAD=${__YAML__BAD:-}

#
# Replacement character for YAML-invalid strings
#
# Characters that fit within ranges defined in $__YAML__BAD will be
# replaced by this one.
#
__YAML__BAD_REPL=${__YAML__BAD_REPL:-�}

yaml__array() {
    #
    # Print an array field named $2 with array of values $3.., nested $1 levels
    #
    local nest=$1; shift    # nesting level
    local name=$1; shift    # field name
    local value             # each value in array
    local sp                # indentation space
    sp=$(__yaml__nest "$nest")
    test $# -eq 0 && echo "    $name: []" && return 0
    echo "$sp$name:"
    for value in "$@"; do
        case $value in
            "") echo "$sp    - ~" ;;
            *)  echo "$sp    - |-4"
                #shellcheck disable=SC2001
                sed "s/^/$sp        /" <<<"$value"
                ;;
        esac
    done | __yaml__strip4yaml
}

yaml__dict() {
    #
    # Print a dict field named $2 and composed of KEY=VALUE pairs $3.., nested $1 levels
    #
    local nest=$1; shift    # nesting level
    local name=$1; shift    # field name
    local pair              # each KEY=VALUE pair
    local key               # each key in dict
    local value             # each value in dict
    local sp                # indentation space
    sp=$(__yaml__nest "$nest")
    test $# -eq 0 && echo "$sp$name: {}" && return 0
    echo "$sp$name:"
    for pair in "$@"; do
        key=${pair%%=*}; value=${pair#$key=}
        case $value in
            "") echo "$sp    $key: ~" ;;
            *)
                echo "$sp    $key: |-4"
                #shellcheck disable=SC2001
                sed "s/^/$sp        /" <<<"$value"
                ;;
        esac
    done | __yaml__strip4yaml
}

yaml__scalar() {
    #
    # Print a scalar field named $2 with value $3, nested $1 levels
    #
    local nest=$1; shift    # nesting level
    local name=$1           # field name
    local value=$2          # scalar value
    local sp                # indentation space
    sp=$(__yaml__nest "$nest")
    case $value in
        "") echo "$sp$name: ~" ;;
        *)
            echo "$sp$name: |-4"
            #shellcheck disable=SC2001
            sed "s/^/$sp    /" <<<"$value"
            ;;
    esac | __yaml__strip4yaml
}

__yaml__nest() {
    #
    # Print $1 x 4 of spaces
    #
    local n=$1
    while test "$n" -gt 0; do
        echo -n '    '
        (( n-- ))
    done
}

__yaml__valid_pair() {
    #
    # True if $1 is a valid K=V pair
    #
    local tainted=$1    # checked data
    grep -qx '\([tor][.]\)\?[[:alpha:]_][[:alnum:]_]*=.*' <<<"$tainted"
}

__yaml__strip4yaml() {
    #
    # Replace yaml-invalid characters
    #
    perl -CS -Mutf8 -pe "tr/$__YAML__BAD/$__YAML__BAD_REPL/"
}

__shellfu_yaml__init() {
    #
    # Initialize 'yaml' module
    #
    if test -z "$__YAML__BAD"; then
        #
        # Yaml won't allow all characters:
        #
        # > [...] The allowed character range explicitly excludes the C0 control
        # > block #x0-#x1F (except for TAB #x9, LF #xA, and CR #xD which are
        # > allowed), DEL #x7F, the C1 control block #x80-#x9F (except for NEL
        # > #x85 which is allowed), the surrogate block #xD800-#xDFFF, #xFFFE,
        # > and #xFFFF.
        #
        # Due to bug in some versions of PyYAML, also replace several extra
        # chars; that is, everything above U+10000.
        #
        __YAML__BAD+='\N{U+0}-\N{U+8}\N{U+B}\N{U+C}\N{U+E}-\N{U+1F}'    # C0 with some gaps
        __YAML__BAD+='\N{U+7F}'                                         # DEL alone
        __YAML__BAD+='\N{U+80}-\N{U+84}\N{U+86}-\N{U+9F}'               # C1 with NEL gap
        # __YAML__BAD+='\N{U+D800}-\N{U+DFFF}'                          # surrogates
        # __YAML__BAD+='\N{U+FFFE}-\N{U+FFFF}'                          # 0xFFFE and 0xFFFF
        #FIXME: for some reasons perl complains about these ^^
        __YAML__BAD+='\N{U+10000}-\N{U+FFFFFF}'               # because of buggy parser(s)
        #FIXME: remove after https://bugzilla.redhat.com/show_bug.cgi?id=1677421
    fi
}


#shellfu module-version=0.0.1
